@@ -6,7 +6,7 @@
 ;; Keywords: tools, wp
 ;; URL: https://github.com/astoff/jit-spell
 ;; Package-Requires: ((emacs "27.1") (compat "29.1"))
-;; Version: 0.1
+;; Version: 0.2
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -42,14 +42,14 @@
 ;;     (define-key jit-spell-mode-map (kbd "C-;") 'jit-spell-correct-word))
 ;;
 ;; jit-spell relies on the `ispell' library to pick a spell checking
-;; program and dictionaries.  Try `M-x customize-group ispell' to see
-;; a listing of all possible settings.
+;; program and dictionaries.  Try `M-x customize-group RET ispell RET'
+;; to see a listing of all possible settings.
 
 ;;; Code:
 
+(require 'cl-lib)
 (require 'compat)
 (require 'ispell)
-(eval-when-compile (require 'subr-x))
 
 (defgroup jit-spell nil
   "Check spelling as you type."
@@ -69,31 +69,58 @@
   :type 'number)
 
 (defcustom jit-spell-ignored-faces
-  '(font-latex-math-face
-    font-latex-sedate-face
-    message-header-name)
+  '(message-header-name)
   "Faces jit-spell should ignore."
   :type '(repeat face))
 
+(defcustom jit-spell-tex-ignored-faces
+  '(font-latex-math-face
+    font-latex-verbatim-face
+    font-latex-sedate-face
+    font-lock-function-name-face
+    font-lock-keyword-face
+    font-lock-variable-name-face
+    tex-math
+    tex-verbatim)
+  "Faces jit-spell should ignore in TeX and derived modes.
+You can modify this variable buffer locally, say in a mode hook,
+but this must be done before activating `jit-spell-mode.'"
+  :type '(repeat face))
+
 (defcustom jit-spell-prog-mode-faces
   '(font-lock-comment-face
     font-lock-doc-face
     font-lock-string-face)
-  "Faces jit-spell should check in modes derived from `prog-mode'."
+  "Faces jit-spell should check in modes derived from `prog-mode'.
+You can modify this variable buffer locally, say in a mode hook,
+but this must done before activating `jit-spell-mode.'"
   :type '(repeat face))
 
-(defvar jit-spell-delayed-commands
-  '(backward-delete-char-untabify
-    delete-backward-char
-    self-insert-command)
-  "List of commands with delayed spell checking.
-Wait for `jit-spell-current-word-delay' seconds before
-highlighting a misspelling at point after one of these commands.")
+(defcustom jit-spell-use-apostrophe-hack 'auto
+  "Whether to work around Hunspell's issue parsing apostrophes.
+In some languages, Hunspell always considers the apostrophe
+character (a.k.a. straight quote) part of the word, which leads
+to false positives when it is used as a quotation mark."
+  :type '(choice (const :tag "Decide automatically" auto)
+                 (const :tag "Yes" t)
+                 (const :tag "No" nil)))
 
-(defvar jit-spell-ignored-p #'jit-spell--default-ignored-p
+(defvar jit-spell-delayed-commands nil)
+(make-obsolete-variable 'jit-spell-delayed-commands "Not necessary anymore" "0.2")
+
+(defvar jit-spell--ignored-p #'jit-spell--default-ignored-p
   "Predicate satisfied by words to ignore.
 It should be a function taking two arguments, the start and end
 positions of the word.")
+(make-obsolete-variable 'jit-spell-ignored-p 'jit-spell--ignored-p "0.2")
+
+(defvar jit-spell--filter-region #'jit-spell--filter-region
+  "Function to extract regions of interest from the buffer.
+It receives the start and end positions of a region as argument,
+and should return a list of cons cells representing subregions.
+
+This function is called inside a `save-excursion' form and can
+move the point with impunity.")
 
 (defvar jit-spell--process-pool nil
   "A collection of ispell processes for `jit-spell-mode'.")
@@ -104,6 +131,9 @@ positions of the word.")
 (defvar-local jit-spell--recheck-timer nil
   "Timer to debounce recheck requests.")
 
+(defvar-local jit-spell--local-words nil
+  "A list of words accepted temporarily in this buffer.")
+
 ;;; Mode-specific support
 
 (defun jit-spell--default-ignored-p (start end)
@@ -115,7 +145,7 @@ positions of the word.")
                       face)
           (memq face jit-spell-ignored-faces)))
       (member (buffer-substring-no-properties start end)
-              ispell-buffer-session-localwords)))
+              jit-spell--local-words)))
 
 (defun jit-spell--prog-ignored-p (start _end)
   "Additional ignore predicate for `prog-mode'."
@@ -125,8 +155,6 @@ positions of the word.")
                        face)
            (memq face jit-spell-prog-mode-faces)))))
 
-;; TODO: jit-spell--org-ignored-p, etc.
-
 ;;; Overlays
 
 (put 'jit-spell 'evaporate t)
@@ -154,11 +182,11 @@ character offset from START, and a list of corrections."
   (with-current-buffer buffer
     (with-silent-modifications
       (remove-list-of-text-properties start end '(jit-spell-pending))
-      (remove-overlays start end 'category 'jit-spell)
+      (jit-spell--remove-overlays start end)
       (pcase-dolist (`(,word ,offset ,corrections) misspellings)
         (let* ((wstart (+ start offset -1))
                (wend (+ wstart (length word))))
-          (unless (funcall jit-spell-ignored-p wstart wend)
+          (unless (funcall jit-spell--ignored-p wstart wend)
             (jit-spell--make-overlay wstart wend corrections)))))))
 
 (defun jit-spell--overlay-at (pos)
@@ -183,6 +211,19 @@ character offset from START, and a list of corrections."
             (unless (< 0 i)
               (throw 'jit-spell ov))))))))
 
+(defun jit-spell--remove-overlays (start end &optional gaps)
+  "Remove all `jit-spell' overlays between START and END, skipping GAPS.
+GAPS must be an ordered list of conses, with the intervals closer
+to END coming first."
+  (pcase-dolist (`(,i . ,j) gaps)
+    (dolist (ov (overlays-in j end))
+      (when (eq 'jit-spell (overlay-get ov 'category))
+        (delete-overlay ov)))
+    (setq end i))
+  (dolist (ov (overlays-in start end))
+    (when (eq 'jit-spell (overlay-get ov 'category))
+      (delete-overlay ov))))
+
 (defun jit-spell--apply-correction (ov text)
   "Replace region spanned by OV with TEXT."
   (save-excursion
@@ -206,11 +247,11 @@ It can also be bound to a mouse click to pop up the menu."
                             (vector corr (lambda () (interactive)
                                            (jit-spell--apply-correction ov corr)))))
       (easy-menu-add-item map nil `["Save to Dictionary"
-                                    (jit-spell-accept-word ,word 'dict)])
+                                    (jit-spell--accept-word ,word 'dict)])
       (easy-menu-add-item map nil `["Save to Buffer"
-                                    (jit-spell-accept-word ,word 'buffer)])
+                                    (jit-spell--accept-word ,word 'buffer)])
       (easy-menu-add-item map nil `["Accept for Session"
-                                    (jit-spell-accept-word ,word 'session)])
+                                    (jit-spell--accept-word ,word 'session)])
       (unless menu (popup-menu map)))
     menu))
 
@@ -222,28 +263,39 @@ It can also be bound to a mouse click to pop up the menu."
      (overlay-put ov 'face face)
      (setq jit-spell--hidden-overlay nil))))
 
-(defun jit-spell--unfontify (&optional start end)
-  "Remove overlays and forget checking status from START to END (or whole buffer)."
-  (save-restriction
-    (widen)
+(defun jit-spell--unfontify (&optional start end lax)
+  "Remove overlays and forget checking status from START to END (or whole buffer).
+Force refontification of the region, unless LAX is non-nil."
+  (without-restriction
     (setq start (or start (point-min)))
     (setq end (or end (point-max)))
-    (remove-overlays start end 'category 'jit-spell)
-    (remove-list-of-text-properties start end '(jit-spell-pending))))
+    (jit-spell--remove-overlays start end)
+    (remove-list-of-text-properties start end '(jit-spell-pending))
+    (unless lax (jit-lock-refontify start end))))
 
 ;;; Subprocess communication
 
 (defun jit-spell--process-parameters ()
-  "Return a list of parameters for this buffer's ispell process.
-Buffers where this produces `equal' results will share their
-ispell process."
+  "Return a list of parameters for this buffer's ispell process."
   (list ispell-program-name
-        ispell-current-dictionary
-        ispell-current-personal-dictionary
+        (or ispell-local-dictionary ispell-dictionary)
+        (or ispell-local-pdict ispell-personal-dictionary)
         ispell-extra-args))
 
 (defun jit-spell--get-process ()
-  "Get an ispell process for the current buffer."
+  "Get an ispell process for the current buffer.
+Buffers where `jit-spell--process-parameters' returns the `equal'
+results share their ispell process.
+
+The process plist includes the following properties:
+
+`jit-spell--current request': a list of the form
+    (BUFFER TICK START END)
+  where TICK is used to keep track of the timeliness of the response.
+  Note that the process receives a single request at a time.
+
+`jit-spell--requests': a FIFO queue with elements in the same
+  form as above."
   (let* ((params (jit-spell--process-parameters))
          (proc (plist-get jit-spell--process-pool params #'equal)))
     (if (process-live-p proc)
@@ -251,6 +303,7 @@ ispell process."
       (unless ispell-async-processp
         (error "`jit-spell-mode' requires `ispell-async-processp'"))
       (ispell-set-spellchecker-params)
+      (ispell-internal-change-dictionary)
       (setq proc (ispell-start-process))
       (set-process-query-on-exit-flag proc nil)
       (setq jit-spell--process-pool
@@ -300,15 +353,14 @@ ispell process."
                  (jit-spell--make-overlays buffer start end misspellings))))
         (delete-region (point-min) (point-max))
         ;; Send next request to ispell process, if applicable
-        (let (request)
-          (while (and (setq request (pop (process-get proc 'jit-spell--requests)))
-                      (pcase-let ((`(,buffer ,tick) request))
-                        (not (and (buffer-live-p buffer)
-                                  (eq tick (buffer-chars-modified-tick buffer))))))
+        (catch 'jit-spell--done
+          (while-let ((request (pop (process-get proc 'jit-spell--requests)))
+                      (buffer (car request)))
             (when (buffer-live-p buffer)
-              (jit-spell--schedule-pending-checks (car request))))
-          (when request
-            (jit-spell--send-request proc request)))))))
+              (if (not (eq (cadr request) (buffer-chars-modified-tick buffer)))
+                  (jit-spell--schedule-pending-checks buffer)
+                (jit-spell--send-request proc request)
+                (throw 'jit-spell--done nil)))))))))
 
 (defun jit-spell--send-request (proc request)
   "Send REQUEST to ispell process PROC."
@@ -316,28 +368,17 @@ ispell process."
   (pcase-let ((`(,buffer _ ,start ,end) request))
     (with-current-buffer buffer
       (let ((text (buffer-substring-no-properties start end)))
-                                        ;TODO: allow custom
-                                        ;buffer-substring functions
-                                        ;e.g. to work around
-                                        ;hunspell's apostrophe issue.
-        ;; Redact control characters in text
-        (dotimes (i (length text))
-          (when (< (aref text i) ?\s)
-            (aset text i ?\s)))
         (process-send-string proc "^")
         (process-send-string proc text)
         (process-send-string proc "\n")))))
 
 (defun jit-spell--schedule-pending-checks (buffer)
   "Schedule a call to `jit-spell--check-pending-regions' in BUFFER."
-  (when (buffer-live-p buffer)
-    (when (bound-and-true-p jit-spell--debug)
-      (message "Scheduled recheck"))
-    (with-current-buffer buffer
-      (when (timerp jit-spell--recheck-timer)
-        (cancel-timer jit-spell--recheck-timer))
-      (setq jit-spell--recheck-timer
-            (run-with-idle-timer 0.1 nil #'jit-spell--check-pending-regions buffer)))))
+  (with-current-buffer buffer
+    (when (timerp jit-spell--recheck-timer)
+      (cancel-timer jit-spell--recheck-timer))
+    (setq jit-spell--recheck-timer
+          (run-with-idle-timer 0.1 nil #'jit-spell--check-pending-regions buffer))))
 
 (defun jit-spell--check-pending-regions (buffer)
   "Enqueue spell check requests for all pending regions of BUFFER."
@@ -356,54 +397,95 @@ ispell process."
                                  (pop (process-get proc 'jit-spell--requests)))))
           (jit-spell--send-request proc request))))))
 
+(defun jit-spell--filter-region (start end)
+  "Return a list of subregions without control characters between START and END."
+  (goto-char start)
+  (let (regions)
+    (while (re-search-forward (rx (+ print)) end t)
+      (push (cons (match-beginning 0) (match-end 0)) regions))
+    regions))
+
+(defun jit-spell--has-face-p (faces v)
+  "Non-nil if V, a face or list of faces, includes any of the FACES."
+  (if (listp v)
+      (seq-some (lambda (f) (memq f faces)) v)
+    (memq v faces)))
+
+(defun jit-spell--refine-by-face (faces &optional only)
+  "Return a function to refine a list of regions based on its faces.
+If ONLY is nil, regions containing any of the FACES are excluded.
+Otherwise, only such regions are kept."
+  (let ((pred (if only
+                  #'jit-spell--has-face-p
+                (lambda (faces v) (not (jit-spell--has-face-p faces v))))))
+    (lambda (regions)
+      (mapcan
+       (pcase-lambda (`(,i . ,limit)) ;; Refine one region
+         (let (result)
+           (with-restriction i limit
+             (goto-char i)
+             (while-let ((prop (text-property-search-forward 'face faces pred)))
+               (push `(,(prop-match-beginning prop) . ,(prop-match-end prop))
+                     result)))
+           (jit-spell--remove-overlays i limit result)
+           result))
+       regions))))
+
+(defun jit-spell--apostrophe-hack (regions)
+  "Refine REGIONS to work around Hunspell's apostrophe issue."
+  (mapcan
+   (pcase-lambda (`(,i . ,limit)) ;; Refine one region
+     (let (result)
+       (goto-char i)
+       (while (re-search-forward
+               (rx (or (seq (or bol (not alpha)) (group ?') alpha)
+                       (seq alpha (group ?') (or (not alpha) eol))))
+               limit t)
+         (backward-char)
+         (let ((j (or (match-end 1) (match-beginning 2))))
+           (push (cons i j) result)
+           (setq i j)))
+       (push (cons i limit) result)))
+   regions))
+
 (defun jit-spell--check-region (start end)
   "Enqueue a spell check request for region between START and END.
 This is intended to be a member of `jit-lock-functions'."
-  (save-excursion ;; Extend region to include whole words
+  (save-excursion
+    ;; Extend region to include whole words
     (goto-char start)
     (setq start (if (re-search-backward "\\s-" nil t) (match-end 0) (point-min)))
     (goto-char end)
-    (setq end (or (re-search-forward "\\s-" nil t) (point-max))))
-  (put-text-property start end 'jit-spell-pending t)
-  (let ((proc (jit-spell--get-process))
-        (request (list (current-buffer) (buffer-chars-modified-tick) start end)))
-    (if (process-get proc 'jit-spell--current-request)
-        (push request (process-get proc 'jit-spell--requests))
-      (jit-spell--send-request proc request)))
-  `(jit-lock-bounds ,start . ,end))
+    (setq end (or (re-search-forward "\\s-" nil t) (point-max)))
+    (let ((proc (jit-spell--get-process))
+          (buffer (current-buffer))
+          (tick (buffer-chars-modified-tick)))
+      (pcase-dolist (`(,i . ,j) (funcall jit-spell--filter-region start end))
+        (put-text-property i j 'jit-spell-pending t)
+        (let ((request (list buffer tick i j)))
+          (if (process-get proc 'jit-spell--current-request)
+              (push request (process-get proc 'jit-spell--requests))
+            (jit-spell--send-request proc request)))))
+    `(jit-lock-bounds ,start . ,end)))
 
 ;;; Interactive commands and major mode
 
-(defun jit-spell-accept-word (word where)
+(defun jit-spell--accept-word (word where)
   "Accept spelling of WORD.
 WHERE can be `dict' (save in personal dictionary), `buffer' (save
 as a local word at the end of the buffer), `session' (accept
 temporarily and in this buffer only), or `query' (ask for one of
 the above)."
-  (interactive
-   (list (completing-read
-          "Add word: "
-          (thread-last
-            (overlays-in (window-start) (window-end))
-            (seq-sort (lambda (o1 o2) (< (overlay-start o1)
-                                         (overlay-start o2))))
-            (seq-keep (lambda (ov)
-                        (when (eq (overlay-get ov 'category) 'jit-spell)
-                          (buffer-substring-no-properties
-                           (overlay-start ov) (overlay-end ov))))))
-          nil nil nil nil
-          (thing-at-point 'word))
-         'query))
   (pcase-exhaustive where
     ('session (when ispell-buffer-local-name
                 (setq ispell-buffer-local-name (buffer-name)))
-              (cl-pushnew word ispell-buffer-session-localwords
+              (cl-pushnew word jit-spell--local-words
                           :test #'equal))
     ('buffer (ispell-add-per-file-word-list word)
-             (jit-spell-accept-word word 'session))
+             (jit-spell--accept-word word 'session))
     ('dict (process-send-string (jit-spell--get-process)
                                 (format "*%s\n#\n" word)))
-    ('query (jit-spell-accept-word
+    ('query (jit-spell--accept-word
              word
              (pcase (read-multiple-choice (substitute-quotes
                                            (format "Add `%s' to" word))
@@ -414,6 +496,7 @@ the above)."
                (`(?b ,_) 'buffer)
                (`(?s ,_) 'session)))))
   (jit-lock-refontify))
+(define-obsolete-function-alias 'jit-spell-accept-word 'jit-spell--accept-word "0.2")
 
 (defun jit-spell-correct-word--next (arg)
   "Perform a spooky action at a distance."
@@ -456,7 +539,7 @@ again moves to the next misspelling."
       ((and count (pred numberp))
        (jit-spell-correct-word count (overlay-start ov)))
       ((and corr (rx bos ?@ (* space) (? (group (+ nonl)))))
-       (jit-spell-accept-word (or (match-string 1 corr) word) 'query))
+       (jit-spell--accept-word (or (match-string 1 corr) word) 'query))
       (corr (jit-spell--apply-correction ov corr)))))
 
 (defalias 'jit-spell-change-dictionary 'ispell-change-dictionary) ;For discoverability
@@ -468,13 +551,7 @@ again moves to the next misspelling."
     (while (search-forward ispell-words-keyword nil t)
       (let ((limit (pos-eol)))
 	(while (re-search-forward "\\s-*\\(\\S-+\\)" limit t)
-          (jit-spell-accept-word (match-string-no-properties 1) 'session))))))
-
-(defun jit-spell--pre-command-hook ()
-  "Pre-command hook for `jit-spell-mode'."
-  (when (and jit-spell--hidden-overlay
-             (not (memq this-command jit-spell-delayed-commands)))
-    (jit-spell--unhide-overlay)))
+          (jit-spell--accept-word (match-string-no-properties 1) 'session))))))
 
 (defvar-keymap jit-spell-mode-map :doc "Keymap for `jit-spell-mode'.")
 
@@ -494,28 +571,43 @@ again moves to the next misspelling."
                                    (mouse-1 . ispell-change-dictionary)))))
   (cond
    (jit-spell-mode
+    ;; Major mode support
     (cond
      ((derived-mode-p 'prog-mode)
-      (add-function :before-until (local 'jit-spell-ignored-p)
-                    #'jit-spell--prog-ignored-p))
-     ((derived-mode-p 'org-mode)
-      (setq-local jit-spell-delayed-commands
-                  (append '(org-delete-backward-char org-self-insert-command)
-                          jit-spell-delayed-commands))))
+      (add-function :filter-return (local 'jit-spell--filter-region)
+                    (jit-spell--refine-by-face jit-spell-prog-mode-faces t)))
+     ((derived-mode-p 'tex-mode)
+      (add-function :filter-return (local 'jit-spell--filter-region)
+                    (jit-spell--refine-by-face jit-spell-tex-ignored-faces))))
+    ;; Generic ignore predicate
+    (when-let ((pred (or (bound-and-true-p flyspell-generic-check-word-predicate)
+                         (get major-mode 'flyspell-mode-predicate))))
+      (add-function :after-until (local 'jit-spell--ignored-p)
+                    (lambda (_start end) (ignore-errors
+                                           (save-excursion
+                                             (goto-char end)
+                                             (not (funcall pred)))))))
+    ;; Hunspell workaround
+    (when (if (eq 'auto jit-spell-use-apostrophe-hack)
+              ispell-really-hunspell
+            jit-spell-use-apostrophe-hack)
+      (add-function :filter-return (local 'jit-spell--filter-region)
+                    #'jit-spell--apostrophe-hack))
+    ;; Buffer setup
     (jit-spell--read-local-words)
     (add-hook 'ispell-change-dictionary-hook 'jit-spell--unfontify nil t)
     (add-hook 'context-menu-functions 'jit-spell--context-menu nil t)
-    (add-hook 'pre-command-hook #'jit-spell--pre-command-hook nil t)
+    ;; Font-lock needs to run first
+    (add-hook 'jit-lock-functions #'jit-spell--check-region 10 t)
     (jit-lock-register #'jit-spell--check-region))
    (t
     (jit-lock-unregister #'jit-spell--check-region)
-    (remove-hook 'pre-command-hook #'jit-spell--pre-command-hook)
     (remove-hook 'context-menu-functions 'jit-spell--context-menu t)
     (remove-hook 'ispell-change-dictionary-hook 'jit-spell--unfontify t)
-    (kill-local-variable 'ispell-buffer-session-localwords)
-    (kill-local-variable 'jit-spell-delayed-commands)
-    (kill-local-variable 'jit-spell-ignored-p)))
-  (jit-spell--unfontify))
+    (kill-local-variable 'jit-spell--local-words)
+    (kill-local-variable 'jit-spell--filter-region)
+    (kill-local-variable 'jit-spell--ignored-p)
+    (jit-spell--unfontify nil nil t))))
 
 ;; Don't litter M-x
 (dolist (sym '(jit-spell--context-menu
@@ -523,12 +615,12 @@ again moves to the next misspelling."
   (put sym 'completion-predicate #'ignore))
 
 (dolist (sym '(jit-spell-change-dictionary
-               jit-spell-correct-word
-               jit-spell-accept-word))
+               jit-spell-correct-word))
   (put sym 'completion-predicate (lambda (_ buffer)
                                    (buffer-local-value 'jit-spell-mode
                                                        buffer))))
 
 (provide 'jit-spell)
 
+; LocalWords:  jit ispell
 ;;; jit-spell.el ends here
